/*  Copyright (C) MOXA Inc. All rights reserved.

    This software is distributed under the terms of the
    MOXA License.  See the file COPYING-MOXA for details.
*/

/*
    socketmode.c

    Routines to support NPort serial port control.

    2007-04-30	Kevin Chu
		new release
    2008-05-30	Albert Yu
		add IPv6 support
*/

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#ifndef	WIN32
	#include	<sys/types.h>
	#include	<sys/socket.h>
	#include	<netinet/in.h>
	#include	<netdb.h>
	#include	<errno.h>
	#include	<unistd.h>
	#include	<fcntl.h>
	#include	<termios.h>
	#include	<time.h>
	#include	<sys/time.h>
#endif
#ifdef	SSL_ON
	#include	<openssl/ssl.h>	
#endif

#include	"socketmode.h"

#ifndef INADDR_NONE			// for Solaris
#define INADDR_NONE (-1)
#endif

/*
 *	Local static functions:
 */
static int	SioOpen(int af, void *ipaddr, int tcpport);
static int	SioWaitAck(SOCKET fd, char cmd);
static int	SioCmdportFd(SOCKET fd);
static int	SioAddFdtbl(SOCKET fd, SOCKET cmdfd);
static int	SioDelFdtbl(SOCKET fd);

#ifdef	SSL_ON
	static void ssl_init(void);
	static SSL* SioportSSL(SOCKET fd);
	static int SioAddSSLtbl(SOCKET fd, SSL	*pssl);
	static int SioDelSSLtbl(SOCKET fd);
	SSL_CTX *sslc_ctx;
#endif	

struct aspp_fds {
	SOCKET	fd_data;
	SOCKET	fd_cmd;
#ifdef	SSL_ON
	SSL	*ssl;
#endif	
};

static struct aspp_fds	asfds[DFD_TABLE_SIZE];

/******************************************************************************
 *  Initial data table using of Socket mode functions                         *
 *  This function must be called once before calling any other ssio_open,     *
 *  ssio_close.., etc.                                                        *
 *                                                                            *
 *  Arg:  none                                                                *
 ******************************************************************************/
int	ssio_init(void)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ ) {
		asfds[i].fd_data = -1;
		asfds[i].fd_cmd = -1;
#ifdef	SSL_ON	
		asfds[i].ssl = NULL;
#endif				
	}
	
#ifdef	SSL_ON
	ssl_init();
#endif		
	return(SSIO_OK);
}

/******************************************************************************
 *  open NPort serial port                                                    *
 *                                                                            *
 *  Args:  char  *server_ip -  IP address of NPort                            *
 *            e.g. 192.168.2.1                                                *
 *         int  cmdport - TCP port number of NPort for command tx/rx          *
 *            number range from 1 to 65535.                                   *
 *         int  dataport - TCP port number of NPort for data tx/rx            *
 *            number range from 1 to 65535.                                   *
 *         int  ssl_enable - enable SSL or not                                *
 *            0 : disable SSL, 1 : enable SSL                                 * 
 *                                                                            *	
 *  Return: >= 0    : file handle ID (open success).                          *
 *             -1   : No such port or port not opened.                        *
 *             -2   : Bad parameter                                           * 	
 *             -200 : Socket initialize error                                 *	
 *             -201 : Max connection reached                                  *	
 *             -300 : SSL initialize error                                    *	
 *             -301 : SSL handshake fail                                      *	
 ******************************************************************************/
int	ssio_open(char *server_ip, int cmdport, int dataport, int ssl_enable)
{
	int af = AF_INET;
	SOCKET	fd, fd2;
	unsigned char ipaddr[16];
#ifdef	SSL_ON
	SSL	*pssl;
#endif

	if ( cmdport < 1 || cmdport > 65535 )
		return(SSIO_BADPORT);
	
	if ( dataport < 1 || dataport > 65535 )
		return(SSIO_BADPORT);
	
	if (cmdport == dataport)
		return(SSIO_BADPORT);

#ifdef IPV6_ON
	{
	struct addrinfo aiHints;
	struct addrinfo *aiList = NULL;

	memset(&aiHints, 0, sizeof(aiHints));
	aiHints.ai_family = PF_UNSPEC;
	if ( getaddrinfo(server_ip, NULL, &aiHints, &aiList) != 0 )
	{
		printf("getaddrinfo() failed.\n");
		return(SSIO_BADPARM);
	}
	af = aiList->ai_family;
	if ( aiList->ai_family == AF_INET6 )
		memcpy(ipaddr, ((struct sockaddr_in6*)(aiList->ai_addr))->sin6_addr.s6_addr, 16);
	else if ( aiList->ai_family == AF_INET )
		memcpy(ipaddr, &(((struct sockaddr_in*)(aiList->ai_addr))->sin_addr.s_addr), 4);
	freeaddrinfo(aiList);
	}
#else
	{
	unsigned long ip4;

	if( (ip4 = inet_addr(server_ip)) == INADDR_NONE )
		return(SSIO_BADPARM);
	memcpy(ipaddr, &ip4, 4);
	}
#endif

	if ( (fd = SioOpen(af, ipaddr, dataport)) < 0 )
		return(SSIO_SOCK_INIT_FAIL);
	if ( (fd2 = SioOpen(af, ipaddr, cmdport)) < 0 )
	{
		closesocket(fd);
		return(SSIO_SOCK_INIT_FAIL);
	}
#ifdef IPV6_ON
	else if (af == AF_INET6)
	{	/* init command port */
		char buf[16];
		memset(buf, 0xff, 16);
		send(fd2, buf, 16, 0);
		SioWaitAck(fd2, DCMD_IPV6_INIT);
	}
#endif
	
	if ( SioAddFdtbl(fd, fd2) < 0 )
	{
		closesocket(fd2);
		closesocket(fd);
		return(SSIO_SOCK_MAX_ERR);
	}
	
#ifdef	SSL_ON
	if (ssl_enable)
	{
		pssl = SSL_new(sslc_ctx);
		if (pssl != NULL)
		{          	              
			if (SSL_set_fd(pssl, fd))
				SSL_set_connect_state(pssl);
		} 
		else
		{
			printf("SSL_new() error.\n");
			return(SSIO_SSL_INIT_FAIL);
		}
  
		if (SSL_connect(pssl) < 0)
		{
			printf("SSL_connect() error.\n");
			SSL_free(pssl);
			closesocket(fd2);
			closesocket(fd);
			SioDelFdtbl(fd);
			return(SSIO_SSL_HANDSHAKE_ERR);
		} 
		SioAddSSLtbl(fd, pssl);
	}
#endif		
	return(fd);
}

/******************************************************************************
 *  close NPort serial port                                                   *
 *                                                                            *
 *  Arg:  SOCKET  fd - file handle ID return from ssio_open()                 *
 ******************************************************************************/
int	ssio_close(SOCKET fd)
{
#ifdef	SSL_ON
	if (SioportSSL(fd) != NULL)
		SioDelSSLtbl(fd);		
#endif	
	
	closesocket(SioCmdportFd(fd));
	closesocket(fd);
	SioDelFdtbl(fd);
	return(SSIO_OK);
}

/******************************************************************************
 *  Set NPort serial port baud rate and character mode                        *
 *                                                                            *
 *  Args:  SOCKET	 fd -	file handle ID return from ssio_open()                *
 *         int  baud - 0: 300      1: 600	     2: 1200                        *
 *                     3: 2400     4: 4800	   5: 7200                        *
 *                     6: 9600     7: 19200    8: 38400                       *
 *                     9: 57600    10: 115200	 11: 230400                     *
 *                     12: 460800  13: 921600	 14: 150                        *
 *                     15: 134     16: 110	   17: 75                         *
 *                     18: 50                                                 *
 *         int  mode - (bit 0,1)  0x00 - 5 data bits                          *
 *                                0x01 - 6 data bits                          *
 *                                0x10 - 7 data bits                          *
 *                                0x11 - 8 data bits                          *
 *                     (bit 2)  0x00 - 1 stop bit                             *
 *                              0x04 - 1.5/2 stop bits                        *
 *                     (bit 3,4,5)  0x00 - none parity                        *
 *                                  0x08 - odd parity                         *
 *                                  0x18 - even parity                        *
 *                                  0x28 - mark parity                        *
 *                                  0x38 - space parity                       *
 *                                                                            *
 *	Return:  0    : O.K.                                                      *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_ioctl(SOCKET fd, int baud, int mode)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_IOCTL;
	buf[1] = 2;
	buf[2] = (unsigned char)baud;
	buf[3] = (unsigned char)mode;
	if ( send(fd, buf, 4, 0) != 4 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}



/******************************************************************************
 *	Set NPort serial port baud rate                                           *
 *                                                                            *
 *  Args:	 SOCKET  fd -	file handle ID return from ssio_open()                *
 *         long  baud -	desired baud rate                                     *
 *                                                                            *
 *	Return:  0    : O.K.                                                      *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_baud(SOCKET fd, long baud)
{
	unsigned char	buf[8];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_BAUDRATE;
	buf[1] = 4;
	*(long *)(&buf[2]) = baud;
	if ( send(fd, buf, 6, 0) != 6 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *	Set NPort serial port flow control                                        *
 *                                                                            *
 *	Args:  SOCKET	 fd -	file handle ID return from ssio_open()                *
 *         int  mode  -	bit 0 : CTS flow control                              *
 *                      bit 1 : RTS flow control                              *
 *                      bit 2 :	Tx XON/XOFF flow control                      *
 *                      bit 3 :	Rx XON/XOFF flow control                      *
 *                      (0 = OFF,  1 = ON)                                    *
 *                                                                            *
 *	Return:  0    : O.K.                                                      *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_flowctrl(SOCKET fd, int mode)
{
	unsigned char	buf[8];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_FLOWCTRL;
	buf[1] = 4;
	if ( mode & 1 )
		buf[2] = 1;
	else
		buf[2] = 0;
	if ( mode & 2 )
		buf[3] = 1;
	else
		buf[3] = 0;
	if ( mode & 4 )
		buf[4] = 1;
	else
		buf[4] = 0;
	if ( mode & 8 )
		buf[5] = 1;
	else
		buf[5] = 0;
	if ( send(fd, buf, 6, 0) != 6 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *	Set NPort serial port line control                                        *
 *                                                                            *
 *	Args:  SOCKET  fd -	file handle ID return from ssio_open()                *
 *         int  mode - bit 0 : DTR output state                               *
 *                     bit 1 : RTS output state                               *
 *                     (0 = OFF,  1 = ON)                                     *
 *                                                                            *
 *	Return:  0    : O.K.                                                      *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_lctrl(SOCKET fd, int mode)
{
	unsigned char	buf[8];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_LINECTRL;
	buf[1] = 2;
	if ( mode & 1 )
		buf[2] = 1;
	else
		buf[2] = 0;
	if ( mode & 2 )
		buf[3] = 1;
	else
		buf[3] = 0;
	if ( send(fd, buf, 4, 0) != 4 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *  Get NPort serial port line status                                         *
 *                                                                            *
 *  Args:  SOCKET	 fd -	file handle ID return from ssio_open()                *
 *                                                                            *
 *  Return:  >=0  : line status:	bit 0 : DSR                                 * 
 *	                              bit 1 : CTS                                 * 
 *                                bit 2 : DCD                                 *
 *                                (1 - ON, 0 - OFF)                           *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_lstatus(SOCKET fd)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_LSTATUS;
	buf[1] = 0;
	if ( send(fd, buf, 2, 0) != 2 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *  Flush input/output buffer data                                            *
 *                                                                            *
 *  Args:	 SOCKET	 fd -	file handle ID return from ssio_open()                *
 *		     int  func -	0 :	flush input buffer                                *
 *					            1 :	flush output buffer	                              *
 *					            2 :	flush input/output buffer                         *
 *                                                                            *
 *  Return:  0    : O.K.                                                      *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_flush(SOCKET fd, int func)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_FLUSH;
	buf[1] = 1;
	buf[2] = func;
	if ( send(fd, buf, 3, 0) != 3 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *  Write data to NPort serial port                                           *
 *                                                                            *
 *  Args:  SOCKET  fd	-	file handle ID return from ssio_open()                *
 *         char  *buf	-	write out data buffer pointer                         *
 *         int  len	-	write out data length                                   *
 *                                                                            *
 *  Return:  >= 0 : write out data length                                     *
 *             -1 : write error                                               *
 ******************************************************************************/
int	ssio_write(SOCKET fd, char *buf, int len)
{
#ifdef	SSL_ON
	SSL *pssl;	
	pssl = SioportSSL(fd);
	if (pssl != NULL)
		return( SSL_write(pssl, buf, len) );
	else
		return( send(fd, buf, len, 0) );	
#else
	return( send(fd, buf, len, 0) );
#endif	
}

/******************************************************************************
 *  Read data from NPort serial port                                          *
 *                                                                            *
 *  Args:	 SOCKET	 fd	-	file handle ID return from ssio_open()                *
 *		     char  *buf	-	read data buffer pointer                              *
 *		     int  len	-	read data buffer length                                 *
 *                                                                            *
 *  Return:  >= 0 : read data length                                          *
 *             -1 : read error                                                *
 ******************************************************************************/
int	ssio_read(SOCKET fd, char *buf, int len)
{
#ifdef	SSL_ON
	SSL *pssl;	
	pssl = SioportSSL(fd);
	if (pssl != NULL)
		return( SSL_read(pssl, buf, len) );
	else	
		return( recv(fd, buf, len, 0) );
#else	
	return( recv(fd, buf, len, 0) );
#endif	
}

/******************************************************************************
 *	Send out a break signal                                                   *
 *                                                                            *
 *	Args:  SOCKET  fd	-	file handle ID return from ssio_open()                *
 *		     int  time - break time in milli-second                             *
 *                                                                            *
 *  Return:  0    : O.K.                                                      *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_break(SOCKET fd, int time)
{
	unsigned char	buf[8];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_BREAK;
	buf[1] = 2;
	*(unsigned short *)&buf[2] = ntohs((unsigned short)time);
	if ( send(fd, buf, 4, 0) != 4 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *  Get output buffer data queue size                                         *
 *                                                                            *
 *  Args:	 SOCKET	 fd	-	file handle ID return from ssio_open()                *
 *                                                                            *
 *  Return: >= 0  : output buffer data queue size                             *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_oqueue(SOCKET fd)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_OQUEUE;
	buf[1] = 0;
	if ( send(fd, buf, 2, 0) != 2 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *  Get input buffer data queue size                                          *
 *                                                                            *
 *  Args:	 SOCKET	 fd	-	file handle ID return from ssio_open()                *
 *                                                                            *
 *  Return: >= 0  : input buffer data queue size                              *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_iqueue(SOCKET fd)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_IQUEUE;
	buf[1] = 0;
	if ( send(fd, buf, 2, 0) != 2 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *  Get break signal count                                                    *
 *                                                                            *
 *  Args:	 SOCKET  fd	-	file handle ID return from ssio_open()                *
 *                                                                            *
 *  Return: >= 0  : break signal count                                        *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_breakcnt(SOCKET fd)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_BREAKCNT;
	buf[1] = 0;
	if ( send(fd, buf, 2, 0) != 2 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *	Get NPort serial port data status                                         *
 *                                                                            *
 *  Args:  SOCKET	 fd -	file handle ID return from ssio_open()                *
 *                                                                            *
 *  Return:  = 0 : no error happened                                          *
 *	         > 0 : data status - bit 0: parity error                          *
 *                               bit 1: framing error                         *
 *                               bit 2: overrun error                         *
 *	                             bit 3: overflow error                        *
 *                               (1 - ON, 0 - OFF)                            *
 *           -203 : Communication with NPort error                            *
 *           -204 : No such command port                                      *
 ******************************************************************************/
int	ssio_data_status(SOCKET fd)
{
	unsigned char	buf[4];

	if ( (fd = SioCmdportFd(fd)) == -1 )
		return(SSIO_NO_CMDPORT);
	buf[0] = DCMD_DSTATUS;
	buf[1] = 0;
	if ( send(fd, buf, 2, 0) != 2 )
		return(SSIO_SOCK_ERR);
	return( SioWaitAck(fd, buf[0]) );
}

/******************************************************************************
 *	TCP Socket protocol control sub-functions                                 *
 ******************************************************************************/
static int SioOpen(int af, void *ipaddr, int tcpport)
{
	Sockaddr	des;
	int	fd, des_len;

	memset(&des, 0, sizeof(des));
	if ( af == AF_INET )
	{
		des.Sin_family = AF_INET;
		des.Sin_addr.s_addr = *(unsigned long*)ipaddr;
		des.Sin_port = htons((unsigned short)tcpport);
		des_len = sizeof(struct sockaddr_in);
	}
#ifdef IPV6_ON
	else if ( af == AF_INET6 )
	{
		des.Sin6_family = AF_INET6;
		memcpy(des.Sin6_addr.s6_addr, ipaddr, 16);
		des.Sin6_port = htons((unsigned short)tcpport);
		des_len = sizeof(struct sockaddr_in6);
	}
#endif
	else
		return -1;

	if ( (fd = socket(af, SOCK_STREAM, 0)) < 0 )
		return(-1);

	if ( connect(fd, (struct sockaddr *)&des, des_len) < 0 )
	{
		closesocket(fd);
		return(-1);
	}

	return(fd);
}

static int SioWaitAck(SOCKET fd, char cmd)
{
	struct timeval	authtime;
	fd_set		readfds;
	char		tmp[20];
	int		len;
	short		n;

	memset(&authtime, 0, sizeof(authtime));
	authtime.tv_usec = 0L;
	authtime.tv_sec = 10;
	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);
	if ( select(fd + 1, &readfds, NULL, NULL, &authtime) < 0 )
		return(SSIO_SOCK_ERR);
	if ( FD_ISSET(fd, &readfds) == 0 )
		return(SSIO_SOCK_ERR);
	len = recv(fd, tmp, 20, 0);
	if ( cmd == DCMD_IQUEUE || cmd == DCMD_OQUEUE ) {
		if ( len < 4 )
			return(SSIO_SOCK_ERR);
		if ( tmp[0] != cmd || tmp[1] != 2 )
			return(SSIO_SOCK_ERR);			
		n = *(short *)(&tmp[2]);
		return( (int)n );
	}
	else if ( cmd == DCMD_BREAKCNT) {
		if ( len < 4 )
			return(SSIO_SOCK_ERR);
		if ( tmp[0] != cmd || tmp[1] != 2 )
			return(SSIO_SOCK_ERR);
		n = tmp[2]*10 + tmp[3];
		return( (int)n );
	}		 
	else if ( cmd == DCMD_LSTATUS ) {
		if ( len < 5 )
			return(SSIO_SOCK_ERR);
		if ( tmp[0] != cmd || tmp[1] != 3 )
			return(SSIO_SOCK_ERR);
		n = 0;
		if ( tmp[2] == 1 )
			n |= 1;
		if ( tmp[3] == 1 )
			n |= 2;
		if ( tmp[4] == 1 )
			n |= 4;
		return( (int)n );
	}
	else if ( cmd == DCMD_DSTATUS ) {
		if ( len < 3 )
			return(SSIO_SOCK_ERR);
		if ( tmp[0] != cmd || tmp[1] != 1 )
			return(SSIO_SOCK_ERR);	
		n = *(char *)(&tmp[2]);
		return( (int)n );	
	}
	else if ( cmd == DCMD_IPV6_INIT ) {
		if (len != 16)
			return(SSIO_SOCK_ERR);
	}
	else {
		if ( len < 3 )
			return(SSIO_SOCK_ERR);
		if ( tmp[0] != cmd || tmp[1] != 'O' || tmp[2] != 'K' )
			return(SSIO_SOCK_ERR);
	}
	return(SSIO_OK);
}

static int SioCmdportFd(SOCKET fd)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ )
		if ( asfds[i].fd_data == fd )
			return(asfds[i].fd_cmd);
	return(-1);
}

static int SioAddFdtbl(SOCKET fd, SOCKET cmdfd)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ ) {
		if ( asfds[i].fd_data == -1 ) {
			asfds[i].fd_data = fd;
			asfds[i].fd_cmd = cmdfd;
			return(0);
		}
	}
	return(-1);
}

static int SioDelFdtbl(SOCKET fd)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ ) {
		if ( asfds[i].fd_data == fd ) {
			asfds[i].fd_data = -1;
			asfds[i].fd_cmd = -1;
			return(0);
		}
	}
	return(-1);
}


#ifdef	SSL_ON
static void ssl_init(void)
{  
  SSLeay_add_ssl_algorithms();
    
#ifdef SSL_VER2
	sslc_ctx = SSL_CTX_new(SSLv2_client_method());
#else
	#ifdef SSL_VER3
		sslc_ctx = SSL_CTX_new(SSLv3_client_method());;	
	#else
		sslc_ctx = SSL_CTX_new(SSLv23_client_method());
	#endif		
#endif

	/* For blocking mode: cause read/write operations to only return after the handshake and successful completion. */
  SSL_CTX_set_mode(sslc_ctx, SSL_MODE_AUTO_RETRY);  
}

static SSL* SioportSSL(SOCKET fd)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ )
		if ( asfds[i].fd_data == fd )
			return(asfds[i].ssl);
	return(NULL);
}

static int SioAddSSLtbl(SOCKET fd, SSL	*pssl)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ ) {
		if ( asfds[i].fd_data == fd ) {
			asfds[i].ssl = pssl;
			return(0);
		}
	}
	return(-1);
}

static int SioDelSSLtbl(SOCKET fd)
{
	int	i;

	for ( i=0; i<DFD_TABLE_SIZE; i++ ) {
		if ( asfds[i].fd_data == fd ) {
			SSL_shutdown(asfds[i].ssl);
			SSL_free(asfds[i].ssl);
			asfds[i].ssl = NULL;
			return(0);
		}
	}
	return(-1);
}
#endif
